import inspect
import os
from pathlib import Path
from typing import BinaryIO, Union, Type, Optional, Any

from serialzy.serializers.base_model import (
    ModelBaseSerializer,
    deserialize_from_file,
    serialize_to_file,
    unpack_model_file
)


class ONNXSerializer(ModelBaseSerializer):
    def __init__(self):
        super().__init__("onnx", __name__)

    def _types_filter(self, typ: Type) -> bool:
        import onnx
        return inspect.isclass(typ) and issubclass(typ, onnx.ModelProto)

    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        import onnx

        if not isinstance(obj, onnx.ModelProto):
            raise ValueError(f"Attempt to serialize not an ONNX model with {__name__}")

        def write_onnx(out_filename: str) -> None:
            with open(out_filename, "wb") as f:
                f.write(obj.SerializeToString())

        serialize_to_file(dest, write_onnx)

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        self._check_types_valid(schema_type, user_type)

        import onnx

        def read_onnx(in_filename: str) -> onnx.ModelProto:
            with open(in_filename, "rb") as f:
                onnx_model = onnx.load(f)
                return onnx_model

        return deserialize_from_file(source, read_onnx)

    def data_format(self) -> str:
        return self.module

    def unpack_model(self, source: BinaryIO, dest_dir: Union[str, os.PathLike]) -> os.PathLike:
        model_path = Path(dest_dir) / f"model.{self.data_format()}"
        unpack_model_file(source, model_path)
        return model_path



import importlib
import logging
import os
import shutil
import tempfile
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, Type, BinaryIO, Any, Union, Callable, Optional, IO

from packaging import version  # type: ignore

from serialzy.types import get_type
from serialzy.api import Schema, VersionBoundary
from serialzy.base import DefaultSchemaSerializerByReference
from serialzy.utils import cached_installed_packages, module_name

__IO_BUFFER_SIZE__ = 65536


# noinspection PyPackageRequirements
def serialize_to_file(dest: BinaryIO, save_to_file) -> None:
    with tempfile.NamedTemporaryFile() as handle:
        save_to_file(handle.name)
        while True:
            data = handle.read(__IO_BUFFER_SIZE__)
            if not data:
                break
            dest.write(data)


def deserialize_from_file(source: BinaryIO, read_from_file) -> Any:
    with tempfile.NamedTemporaryFile() as handle:
        while True:
            data = source.read(__IO_BUFFER_SIZE__)
            if not data:
                break
            handle.write(data)
        handle.flush()
        os.fsync(handle.fileno())
        return read_from_file(handle.name)


def serialize_to_dir(dest: BinaryIO, save_to_dir):
    tmpdir = tempfile.mkdtemp()
    save_to_dir(tmpdir)
    tar_save_path = os.path.join(tmpdir, "model")
    saved_archive = shutil.make_archive(tar_save_path, "tar", tmpdir)
    with open(saved_archive, "rb") as handle:
        while True:
            data = handle.read(__IO_BUFFER_SIZE__)
            if not data:
                break
            dest.write(data)
    shutil.rmtree(tmpdir)


def deserialize_from_dir(source: BinaryIO, read_from_dir) -> Any:
    with tempfile.NamedTemporaryFile(suffix=".tar") as handle:
        while True:
            data = source.read(__IO_BUFFER_SIZE__)
            if not data:
                break
            handle.write(data)
        handle.flush()
        os.fsync(handle.fileno())

        tmpdir = tempfile.mkdtemp()
        shutil.unpack_archive(handle.name, tmpdir, "tar")
        model = read_from_dir(tmpdir)
        shutil.rmtree(tmpdir)
        return model


def unpack_model_file(source: IO[bytes], destination: Path) -> None:
    with destination.open("wb") as handle:
        __unpack_model(source, handle)


def unpack_model_tar(source: IO[bytes], destination: Path) -> None:
    with tempfile.NamedTemporaryFile(suffix=".tar") as handle:
        __unpack_model(source, handle)

        handle.flush()
        os.fsync(handle.fileno())

        shutil.unpack_archive(handle.name, destination, "tar")


def __unpack_model(source: IO[bytes], destination: IO[bytes]) -> None:
    while True:
        data = source.read(__IO_BUFFER_SIZE__)
        if not data:
            break
        destination.write(data)


class ModelBaseSerializer(DefaultSchemaSerializerByReference, ABC):
    META_FILE_NAME = 'meta.json'

    def __init__(self, module: str, logger_name: str):
        self.module = module
        self.logger = logging.getLogger(logger_name)

    @abstractmethod
    def unpack_model(self, source: BinaryIO, dest_dir: Union[str, os.PathLike]) -> os.PathLike:
        """
        Create model file in `dest_dir` and write serialized model (without serialzy metadata) into the file
        :param source: model data is retrieving from source
        :param dest_dir: model data will be layouted in dest_dir
        :return path to model data
        """
        pass

    def available(self) -> bool:
        # noinspection PyBroadException
        try:
            importlib.import_module(self.module)

            return True
        except Exception:
            return False

    def stable(self) -> bool:
        return True

    def meta(self) -> Dict[str, str]:
        m = importlib.import_module(self.module)

        return {self.module: m.__version__}

    def resolve(self, schema: Schema) -> Type:
        typ = super().resolve(schema)
        if self.module not in schema.meta:
            self.logger.warning(f'No {self.module} version in meta')
        elif version.parse(schema.meta[self.module]) > version.parse(cached_installed_packages[self.module]):
            self.logger.warning(f'Installed version of {self.module} {cached_installed_packages[self.module]} '
                                f'is older than used for serialization {schema.meta[self.module]}')
        return typ

    def requirements(self) -> Dict[str, VersionBoundary]:
        return {self.module: VersionBoundary()}

    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        return lambda t: self.__check_module_fits_serializer(t) and self._types_filter(t)

    def _types_filter(self, typ: Type) -> bool:
        return False

    def __check_module_fits_serializer(self, typ: Type) -> bool:
        name = module_name(typ)
        if name == self.module:
            return True

        if hasattr(typ, '__bases__'):
            for base in typ.__bases__:
                base_name = module_name(base)
                if base_name == self.module:
                    return True
        return False

    def serialize_with_meta(self, obj: Any, dest_dir: str, model_file_name: str) -> None:
        """
        :param obj: object to serialize into bytes
        :param dest_dir: files with serialized model and meta are written into destination directory
        :param model_file_name: serialized model's file name
        :return: None
        """
        dest_dir_path = Path(dest_dir)

        if not dest_dir_path.is_dir():
            raise NotADirectoryError(f"given path {dest_dir_path} is not a dir")

        typ = get_type(obj)
        # check that obj type is valid
        self._check_type(typ)

        # write schema header
        meta_file_path = dest_dir_path / ModelBaseSerializer.META_FILE_NAME
        with meta_file_path.open('wb') as dest:
            self._write_schema(typ, dest)

        # write serialized model
        model_file_path = dest_dir_path / model_file_name
        with model_file_path.open('wb') as dest:
            self._serialize(obj, dest)

    def deserialize_with_meta(self, source_dir: str, model_file_name: str, typ: Optional[Type] = None) -> Any:
        """
        :param source_dir: directory with files with serialized model and meta
        :param model_file_name: name of file with serialized model
        :param typ: type of the resulting object, fetched from the meta if None
        :return:
        """
        dest_dir_path = Path(source_dir)

        # read schema header
        meta_file_path = dest_dir_path / ModelBaseSerializer.META_FILE_NAME
        with meta_file_path.open('rb') as source:
            schema_type = self._deserialize_type(source)

        # read serialized data
        model_file_path = dest_dir_path / model_file_name
        with model_file_path.open('rb') as source:
            return self._deserialize(source, schema_type, typ)



import sys
from typing import Any, BinaryIO, Type, Optional, Union, Callable, Dict

from serialzy.api import VersionBoundary, Serializer, Schema, StandardSchemaFormats
from serialzy.version import __version__


class EllipsisSerializer(Serializer):
    ellipsis_data_format = "serialzy_python_ellipsis"

    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        pass

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        self._check_types_valid(schema_type, user_type)
        return schema_type

    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        # noinspection PyTypeChecker
        return lambda x: x == Ellipsis  # type: ignore

    def available(self) -> bool:
        return True

    def stable(self) -> bool:
        return True

    def data_format(self) -> str:
        return self.ellipsis_data_format

    def meta(self) -> Dict[str, str]:
        version = sys.version_info
        return {
            'serialzy': __version__,
            'python': f"{version.major}.{version.minor}.{version.micro}"
        }

    def requirements(self) -> Dict[str, VersionBoundary]:
        return {}

    def schema(self, typ: Type) -> Schema:
        return Schema(
            self.data_format(),
            StandardSchemaFormats.no_schema.name,
            "",
            self.meta(),
        )

    def resolve(self, schema: Schema) -> Type:
        self._validate_schema(schema)
        # noinspection PyTypeChecker
        return Ellipsis  # type: ignore



import copyreg
import sys
from types import TracebackType, CodeType, FunctionType
from typing import Callable, List


def register_exception_serializer_to_pickle():
    copyreg.pickle(TracebackType, __pickle_traceback)


def __pickle_traceback(tb):
    return to_traceback, (from_traceback(tb),)


class Traceback:
    def __init__(self, tb):
        self.co_name = tb.tb_frame.f_code.co_name
        self.tb_lineno = int(tb.tb_lineno)
        self.co_filename = tb.tb_frame.f_code.co_filename
        self.f_globals = {k: v for k, v in tb.tb_frame.f_globals.items() if k in ("__file__", "__name__")}


def from_traceback(tb):
    stack = []
    while tb:
        stack.append(Traceback(tb))
        tb = tb.tb_next
    return stack


def to_traceback(tbs):
    stack: List[Callable] = [ValueError]
    for i, tb in enumerate(reversed(tbs)):
        code = compile("\n" * (tb.tb_lineno - 1) + "raise exception()", tb.co_filename, "exec")
        if hasattr(code, "replace"):
            code = code.replace(co_argcount=0, co_filename=tb.co_filename, co_name=tb.co_name,
                                co_freevars=(), co_cellvars=())
        elif sys.version_info < (3, 8):
            code = CodeType(
                co_argcount=0,
                co_kwonlyargcount=code.co_kwonlyargcount,
                co_nlocals=code.co_nlocals,
                co_stacksize=code.co_stacksize,
                co_coflags=code.co_flags,
                co_code=code.co_code,
                co_consts=code.co_consts,
                co_co_names=code.co_names,
                co_varnames=code.co_varnames,
                co_filename=tb.co_filename,
                co_coname=tb.co_name,
                co_firstlineno=code.co_firstlineno,
                co_lnotab=code.co_lnotab,
                co_freevars=(),
                co_cellvars=(),
            )
        else:
            assert False

        tb.f_globals["exception"] = stack[i]
        func = FunctionType(code, tb.f_globals, tb.co_name)
        stack.append(func)

    try:
        stack[-1]()
    except ValueError:
        tb = sys.exc_info()[2]
        assert tb
        return tb.tb_next



from typing import Any, Type, List, Tuple, Callable, Optional, Dict


class EmptyContent:
    pass


def get_type(obj: Any, type_provider: Optional[Callable[[Any], Type]] = None) -> Type:
    original_type = type(obj) if type_provider is None else type_provider(obj)
    if original_type == list and len(obj) > 0:
        return List[get_type(obj[0], type_provider)]  # type: ignore  # suppose that all elements have the same type
    elif original_type == tuple and 0 < len(obj) <= 20:  # suppose that a human cannot make tuple with more than 20 args
        return Tuple[tuple(get_type(x, type_provider) for x in obj)]  # type: ignore
    elif original_type == tuple and len(obj) > 0 and all(isinstance(x, type(obj[0])) for x in obj):
        return Tuple[get_type(obj[0], type_provider), ...]  # type: ignore
    elif original_type == list and len(obj) == 0:
        return List[EmptyContent]  # type: ignore
    elif original_type == tuple and len(obj) == 0:
        return Tuple[EmptyContent]  # type: ignore
    elif original_type == dict and len(obj) > 0:
        k, v = next(iter(obj.items()))
        return Dict[get_type(k), get_type(v)]  # type: ignore
    elif original_type == dict and len(obj) == 0:
        return Dict[EmptyContent, EmptyContent]  # type: ignore
    else:
        return original_type



import inspect
import json
from os import PathLike
from pathlib import Path
from typing import Type, Any, BinaryIO, Optional, Union

from serialzy.api import Schema
from serialzy.serializers.base_model import (
    ModelBaseSerializer,
    serialize_to_dir,
    deserialize_from_dir,
    unpack_model_tar
)


# noinspection PyPackageRequirements
class TensorflowKerasSerializer(ModelBaseSerializer):
    def __init__(self):
        super().__init__("keras", __name__)

    def unpack_model(self, source: BinaryIO, dest_dir: Union[str, PathLike]) -> PathLike:
        model_path = Path(dest_dir) / "model.savedmodel"
        unpack_model_tar(source, model_path)
        return model_path

    def data_format(self) -> str:
        return "tf_keras"

    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        import tensorflow as tf  # type: ignore
        serialize_to_dir(dest, lambda x: tf.keras.models.save_model(obj, x, save_format='tf'))

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        self._check_types_valid(schema_type, user_type)
        import tensorflow as tf  # type: ignore
        return deserialize_from_dir(source, lambda x: tf.keras.models.load_model(x))

    def _types_filter(self, typ: Type):
        import tensorflow as tf  # type: ignore
        return typ in [tf.keras.Sequential] or (inspect.isclass(typ) and issubclass(typ, tf.keras.Model))


# noinspection PyPackageRequirements
class TensorflowPureSerializer(ModelBaseSerializer):
    def __init__(self):
        super().__init__("tensorflow", __name__)

    def schema(self, typ: Type) -> Schema:
        if self._is_generic_user_object(typ):
            return Schema(
                data_format=self.data_format(),
                schema_format=self.SCHEMA_FORMAT,
                schema_content=json.dumps({
                    "module": self.module,
                    "name": None
                }),
                meta=self.meta()
            )
        return super().schema(typ)

    def unpack_model(self, source: BinaryIO, dest_dir: Union[str, PathLike]) -> PathLike:
        model_path = Path(dest_dir) / "model.savedmodel"
        unpack_model_tar(source, model_path)
        return model_path

    def data_format(self) -> str:
        return "tf_pure"

    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        import tensorflow as tf  # type: ignore
        serialize_to_dir(dest, lambda x: tf.saved_model.save(obj, x))

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        self._check_types_valid(schema_type, user_type)
        import tensorflow as tf  # type: ignore
        return deserialize_from_dir(source, lambda x: tf.saved_model.load(x))

    def _types_filter(self, typ: Type):
        import tensorflow as tf  # type: ignore
        return (
            typ in [tf.train.Checkpoint] or
            (inspect.isclass(typ) and issubclass(typ, tf.Module)) or
            self._is_generic_user_object(typ)
        )

    @staticmethod
    def _is_generic_user_object(typ: Type) -> bool:
        return (
            inspect.isclass(typ) and
            getattr(typ, "__module__", "none") == "tensorflow.python.saved_model.load" and
            getattr(typ, "__name__", "none") == "_UserObject"
        )



from os import PathLike
from pathlib import Path
from typing import BinaryIO, Callable, Type, Union, Any, Optional

from serialzy.errors import SerialzyError
from serialzy.serializers.base_model import (
    ModelBaseSerializer,
    serialize_to_file,
    deserialize_from_file,
    unpack_model_file
)


# noinspection PyPackageRequirements
class CatboostPoolSerializer(ModelBaseSerializer):
    def __init__(self):
        super().__init__("catboost", __name__)

    def unpack_model(self, source: BinaryIO, dest_dir: Union[str, PathLike]) -> PathLike:
        model_path = Path(dest_dir) / "model.bin"
        unpack_model_file(source, model_path)
        return model_path

    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        if not obj.is_quantized():  # type: ignore
            raise SerialzyError('Only quantized pools can be serialized')
        serialize_to_file(dest, lambda x: obj.save(x))

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        self._check_types_valid(schema_type, user_type)
        import catboost
        return deserialize_from_file(source, lambda x: catboost.Pool("quantized://" + x))

    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        import catboost

        return catboost.Pool  # type: ignore

    def data_format(self) -> str:
        return "catboost_quantized_pool"


# noinspection PyPackageRequirements
class CatboostModelSerializer(ModelBaseSerializer):
    def __init__(self):
        super().__init__("catboost", __name__)

    def unpack_model(self, source: BinaryIO, dest_dir: Union[str, PathLike]) -> PathLike:
        model_path = Path(dest_dir) / f"model.{self.data_format()}"
        unpack_model_file(source, model_path)
        return model_path

    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        serialize_to_file(dest, lambda x: obj.save_model(x, format=self.data_format()))

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        self._check_types_valid(schema_type, user_type)
        model = schema_type()
        return deserialize_from_file(source, lambda x: model.load_model(x, format=self.data_format()))

    def _types_filter(self, typ: Type) -> bool:
        import catboost as cb
        return typ in [cb.CatBoostClassifier, cb.CatBoostRegressor, cb.CatBoostRanker]  # type: ignore

    def data_format(self) -> str:
        return "cbm"  # CatBoost binary format https://catboost.ai/en/docs/concepts/python-reference_catboost_save_model



import inspect
from os import PathLike
from pathlib import Path
from typing import BinaryIO, Type, Any, Optional, Union

from serialzy.serializers.base_model import (
    ModelBaseSerializer,
    serialize_to_file,
    deserialize_from_file,
    unpack_model_file
)


# noinspection PyPackageRequirements
class TorchSerializer(ModelBaseSerializer):
    def __init__(self):
        super().__init__("torch", __name__)

    def unpack_model(self, source: BinaryIO, dest_dir: Union[str, PathLike]) -> PathLike:
        model_path = Path(dest_dir) / "model.pt"
        unpack_model_file(source, model_path)
        return model_path

    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        import torch  # type: ignore
        serialize_to_file(dest, lambda x: torch.jit.script(obj).save(x))

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        self._check_types_valid(schema_type, user_type)
        import torch  # type: ignore
        return deserialize_from_file(source, lambda x: torch.jit.load(x))

    def _types_filter(self, typ: Type) -> bool:
        import torch  # type: ignore
        return inspect.isclass(typ) and issubclass(typ, torch.nn.Module)

    def data_format(self) -> str:
        return "pt"



import dataclasses
import importlib
import json
import logging
from abc import ABC
from io import BytesIO
from packaging import version
from typing import Any, BinaryIO, Type, Optional, Dict, cast, Union, Callable, Tuple, List

from serialzy.serializers.ellipsis import EllipsisSerializer
from serialzy.utils import cached_installed_packages
from typing_extensions import get_args, get_origin

from serialzy.api import Serializer, SerializerRegistry, Schema, VersionBoundary
from serialzy.types import EmptyContent, get_type
from serialzy.version import __version__

_LOG = logging.getLogger(__name__)


class SequenceSerializerBase(Serializer, ABC):
    SUPPORTED_TYPES = {list, tuple}
    SCHEMA_FORMAT = "serialzy_sequence_schema"

    def __init__(self, registry: SerializerRegistry):
        self._registry = registry

    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        length = len(obj)
        dest.write(length.to_bytes(length=8, byteorder='little', signed=False))

        for i in range(len(obj)):
            with BytesIO() as handle:
                serializer = cast(Serializer, self._registry.find_serializer_by_type(get_type(obj[i])))
                data_format = serializer.data_format()
                handle.write(len(data_format).to_bytes(length=8, byteorder='little', signed=False))
                handle.write(data_format.encode("utf-8"))

                serializer.serialize(obj[i], handle)
                handle.flush()

                serialized_value = handle.getvalue()
                elem_size = len(serialized_value)
                dest.write(elem_size.to_bytes(length=8, byteorder='little', signed=False))
                dest.write(serialized_value)

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        self._check_types_valid(schema_type, user_type)

        length = int.from_bytes(source.read(8), byteorder='little', signed=False)
        if length == 0:
            return get_origin(schema_type)([])  # type: ignore
        # allow list deserialization by both stable and unstable serializers

        result = list()
        for i in range(length):
            elem_length = int.from_bytes(source.read(8), byteorder='little', signed=False)
            with BytesIO() as handle:
                handle.write(source.read(elem_length))
                handle.flush()
                handle.seek(0)

                data_format_length = int.from_bytes(handle.read(8), byteorder='little', signed=False)
                data_format = handle.read(data_format_length).decode("utf-8")
                serializer = cast(Serializer, self._registry.find_serializer_by_data_format(data_format))
                obj = serializer.deserialize(handle)
                result.append(obj)

        return get_origin(schema_type)(result)  # type: ignore

    def available(self) -> bool:
        return True

    def meta(self) -> Dict[str, str]:
        return {'serialzy': __version__}

    def schema(self, typ: Type) -> Schema:
        args: Tuple[Type, ...] = get_args(typ)
        # noinspection PyProtectedMember,PyUnresolvedReferences
        schema_dict = {
            "origin": {
                "module": typ.__module__,
                "name": typ.__name__ if hasattr(typ, "__name__") else typ._name  # for typing.List
            },
            "args": [dataclasses.asdict(cast(Serializer, self._registry.find_serializer_by_type(arg)).schema(arg)) for
                     arg in args]
        }
        return Schema(self.data_format(), self.SCHEMA_FORMAT, json.dumps(schema_dict), self.meta())

    def resolve(self, schema: Schema) -> Type:
        # do not check data format here to allow list deserialization by both stable and unstable serializers
        if schema.schema_format != self.SCHEMA_FORMAT:
            raise ValueError(f'Invalid schema format {schema.schema_format}')

        if 'serialzy' not in schema.meta:
            _LOG.warning('No serialzy version in meta')
        elif version.parse(schema.meta['serialzy']) > version.parse(cached_installed_packages["serialzy"]):
            _LOG.warning(f'Installed version of serialzy {cached_installed_packages["serialzy"]} '
                         f'is older than used for serialization {schema.meta["serialzy"]}')

        schema_dict = json.loads(schema.schema_content)
        origin = schema_dict["origin"]
        typ = getattr(importlib.import_module(origin["module"]), origin["name"])
        schemas = [Schema(**arg) for arg in schema_dict["args"]]

        if typ == List:
            serializer = cast(Serializer, self._registry.find_serializer_by_data_format(schemas[0].data_format))
            arg_type = serializer.resolve(schemas[0])
            return typ[arg_type]  # type: ignore
        else:  # tuple
            if len(schemas) == 2 and schemas[1].data_format == EllipsisSerializer.ellipsis_data_format:
                serializer = cast(Serializer, self._registry.find_serializer_by_data_format(schemas[0].data_format))
                arg_type = serializer.resolve(schemas[0])
                return typ[arg_type, ...]  # type: ignore
            else:
                typ_args = tuple(
                    cast(Serializer, self._registry.find_serializer_by_data_format(s.data_format)).resolve(s) for s in
                    schemas)
                return typ[typ_args]  # type: ignore

    def requirements(self) -> Dict[str, VersionBoundary]:
        return {}

    def __serializers_from_type_args(self, args: Tuple[Any, ...]) -> List[Serializer]:
        serializers: List[Serializer]
        if (len(args) == 2 and args[1] == Ellipsis) or len(args) == 1:  # tuple with ellipsis or list
            serializers = [cast(Serializer, self._registry.find_serializer_by_type(args[0]))]
        else:
            serializers = [cast(Serializer, self._registry.find_serializer_by_type(arg)) for arg in args]
        return serializers


class SequenceSerializerStable(SequenceSerializerBase):
    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        return lambda t: get_origin(t) in self.SUPPORTED_TYPES and self.__check_arg(get_args(t))

    def stable(self) -> bool:
        return True

    def __check_arg(self, args: Tuple[Any, ...]) -> bool:
        if len(args) == 0:
            return False
        elif args[0] == EmptyContent:
            return True

        for arg in args:
            serializer = self._registry.find_serializer_by_type(arg)
            if serializer is None or not serializer.available() or not serializer.stable():
                return False
        return True

    def data_format(self) -> str:
        return "serialzy_sequence_stable"


class SequenceSerializerUnstable(SequenceSerializerBase):
    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        return lambda t: get_origin(t) in self.SUPPORTED_TYPES and self.__check_arg(get_args(t))

    def stable(self) -> bool:
        return False

    def __check_arg(self, args: Tuple[Any, ...]) -> bool:
        if len(args) == 0:
            return False
        elif args[0] == EmptyContent:
            return True

        for arg in args:
            serializer = self._registry.find_serializer_by_type(arg)
            if serializer is None or not serializer.available():
                return False
        return True

    def data_format(self) -> str:
        return "serialzy_sequence_unstable"
